import json
import re
from collections import OrderedDict

import bleach
from future.backports.urllib.parse import quote


def clean_api_dict(input_value):
    if isinstance(input_value, OrderedDict):
        input_value = to_dict(input_value)

    return convert_keys(clean_api_results(input_value))


def clean_api_results(api_result, **kwargs):
    """
    This function will accept, a list, dictionary, or string.
    If a list or dictionary is provided, it will recursively traverse through it, and run bleach on all string values.
    If a string is provided, it will bleach that string.
    Any other data type is simply returned as-is.

    :param input_dict:
    :return:
    """
    retn = None

    if isinstance(api_result, str):

        retn = bleach.clean(api_result)

    elif isinstance(api_result, dict):
        retn = {}

        for k, v in api_result.items():

            if isinstance(v, str):
                retn[k] = bleach.clean(v)

            else:
                retn[k] = clean_api_results(v)

    elif isinstance(api_result, list):
        retn = []

        for list_item in api_result:
            if isinstance(list_item, str):

                retn.append(bleach.clean(list_item))

            elif isinstance(list_item, dict) or isinstance(list_item, list):

                retn.append(clean_api_results(list_item))

            else:
                retn.append(list_item)

    else:
        retn = api_result

    return retn


def convert_keys(input_value):
    """
    Convert all of the keys in a dict recursively from CamelCase to snake_case.
    Also strips leading and trailing whitespace from string values.
    :param input_value:
    :return:
    """
    retn = None

    if isinstance(input_value, list):
        retn = []
        for list_item in input_value:
            if isinstance(list_item, (dict, list)):
                retn.append(convert_keys(list_item))
            else:
                if isinstance(list_item, str):
                    retn.append(list_item.strip())

                else:
                    retn.append(list_item)

    elif isinstance(input_value, dict):
        retn = dict()
        for k, v in input_value.items():
            new_key_s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", k)
            new_key = re.sub("([a-z0-9])([A-Z])", r"\1_\2", new_key_s).lower()
            if isinstance(v, (dict, list)):
                retn[new_key] = convert_keys(v)

            else:
                if isinstance(v, str):
                    retn[new_key] = v.strip()
                else:
                    retn[new_key] = v

    return retn


def filter_query_characters(value):
    # Convert to a string, and filter anything that has no business being in a URL

    return quote(str(value))


def http_build_query(query_dict):
    """
    Build HTTP Query parameters without urlencoding them further.
    :param query_dict:
    :return:
    """
    query_vals = []
    for key, val in query_dict.items():
        query_vals.append("%s=%s" % (filter_query_characters(key), filter_query_characters(val)))

    return "&".join(query_vals)

def snake_to_camel(value, **kwargs):
    """
    Converts a snake_case key name to a camelCase key name, or vice versa.
    :param value: A string that you want to convert to another string type.
    :param kwargs:
        reverse - Convert from camelCase to snake_case
    :return: a converted string
    """
    do_reverse = kwargs.get("reverse", False)

    parts_ex = "([A-Z])" if do_reverse else "(_[A-Za-z])"
    parts = re.findall(parts_ex, value)

    for part in parts:
        if do_reverse:
            value = value.replace(part, "_%s" % part.lower())

            if value[0] == "_":
                value = value[1:]

        else:
            value = value.replace(part, part.upper().replace("_", ""))

    return value


def to_dict(input_ordered_dict):
    return json.loads(json.dumps(input_ordered_dict))


import datetime
import logging
import os
import pprint
from collections import OrderedDict

import requests
import xmltodict
from future.standard_library import install_aliases

from geo_ez.data_functions import snake_to_camel, http_build_query, clean_api_dict

install_aliases()

logger = logging.getLogger(__name__)


class USPS:
    is_debug = False
    is_logging = False

    today = None

    usps_id = None

    endpoint = "https://secure.shippingapis.com/ShippingAPI.dll"

    def __init__(self, **kwargs):
        self.usps_id = kwargs.get("id", os.environ.get("USPS_ID", False))

        self.is_debug = kwargs.get("debug", os.environ.get("FED_DEBUG", False)) in ["True", True]
        self.is_logging = kwargs.get("logging", os.environ.get("FED_LOGGING", False)) in ["True", True]

        self.today = datetime.datetime.now()

    def debug_msg(self, message, **kwargs):
        pretty = kwargs.get("pretty", False)

        if pretty:
            message = pprint.pformat(message, indent=4)

        if self.is_debug:
            print(message)

        if self.is_logging:
            logger.info(message)

    def usps_call(self, **kwargs):
        api = kwargs.get("api", False)
        data = kwargs.get("data", False)

        resp = {}

        if self.usps_id and api and data:
            api_resp_var = kwargs.get("resp_variable", "%s_response" % snake_to_camel(api, reverse=True))
            self.debug_msg(api_resp_var)

            xml = xmltodict.unparse(OrderedDict(data), full_document=False)
            self.debug_msg(xml)

            endpoint = kwargs.get("endpoint", self.endpoint)

            params = dict(API=api, XML=xml)

            request_url = "%s?%s" % (endpoint, http_build_query(params))

            ret = requests.get(request_url, verify=True)
            self.debug_msg(ret.text)

            # Convert the returned XML to a dict, clean the returned data, and convert all keys to snake case.
            ret_dict = clean_api_dict(xmltodict.parse(ret.text))
            self.debug_msg(ret_dict, pretty=True)

            resp = ret_dict.get(api_resp_var)

        return resp

    def address(self, **kwargs):
        """
        Uses the USPS Address validation API to return a valid address

        NOTE: The USPS API expects Address2 to be the street address, and Address1 to be the suite/apt #,
              so in the code below, we swap the parameters when they are being fed to the API, and when they are
              returned from the API.

        :param kwargs:
            address1, address2, city, state, zip_code, plus_four

        :return:
            A Dict with a valid address:
                {
                    'address1': '',
                    'address2': '',
                    'city': '',
                    'state': 'XX',
                    'zip_code': 'XXXXX',
                    'plus_four': 'XXXX'
                }
        """
        valid_address = None

        address_dict = OrderedDict()
        address_dict["Address1"] = kwargs.get("address1", "")
        address_dict["Address2"] = kwargs.get("address2", "")
        address_dict["City"] = kwargs.get("city", "")
        address_dict["State"] = kwargs.get("state", "")
        address_dict["Zip5"] = kwargs.get("zip_code", "")
        address_dict["Zip4"] = kwargs.get("plus_four", "")

        address_validate_request = OrderedDict()
        address_validate_request["@USERID"] = self.usps_id
        address_validate_request["Address"] = address_dict

        address_validate_dict = {"AddressValidateRequest": address_validate_request}

        resp = self.usps_call(api="Verify", data=address_validate_dict, resp_variable="address_validate_response")

        if "address" in resp:
            ret_address = resp.get("address")
            valid_address = {
                "address1": ret_address.get("address2"),
                "address2": ret_address.get("address1"),
                "city": ret_address.get("city"),
                "state": ret_address.get("state"),
                "zip_code": ret_address.get("zip5"),
                "plus_four": ret_address.get("zip4"),
            }

        return valid_address

    def get_zipcode(self, **kwargs):
        """
        Uses the USPS Zip Code Lookup API to determine a Zip Code based on a partial address.
        :param kwargs:
            address1, address2, city, state, zip_code, plus_four

        :return:
            A Dict with a valid address:
                {
                    'address1': '',
                    'address2': '',
                    'city': '',
                    'state': 'XX',
                    'zip_code': 'XXXXX',
                    'plus_four': 'XXXX'
                }
        """
        valid_address = {"address1": "", "address2": "", "city": "", "state": "XX"}

        address_dict = OrderedDict()
        address_dict["Address1"] = kwargs.get("address2", "")
        address_dict["Address2"] = kwargs.get("address1", "")
        address_dict["City"] = kwargs.get("city", "")
        address_dict["State"] = kwargs.get("state", "")

        address_validate_request = OrderedDict()
        address_validate_request["@USERID"] = self.usps_id
        address_validate_request["Address"] = address_dict

        address_validate_dict = {"ZipCodeLookupRequest": address_validate_request}

        resp = self.usps_call(api="ZipCodeLookup", data=address_validate_dict)

        if "address" in resp:
            ret_address = resp.get("address")
            valid_address = {
                "address1": ret_address.get("address2"),
                "address2": ret_address.get("address1"),
                "city": ret_address.get("city"),
                "state": ret_address.get("state"),
                "zip_code": ret_address.get("zip5"),
                "plus_four": ret_address.get("zip4"),
            }

        return valid_address

    def track(self, tracking_number, **kwargs):
        detailed_response = kwargs.get("details", False)

        if detailed_response:
            req_key = "TrackFieldRequest"

        else:
            req_key = "TrackRequest"

        tracking_dict = OrderedDict(
            {req_key: OrderedDict({"@USERID": self.usps_id, "TrackID": OrderedDict({"@ID": tracking_number})})}
        )

        resp = self.usps_call(api="TrackV2", data=tracking_dict, resp_variable="track_response")

        tracking_info = resp.get("track_info")

        tracking_data = {
            "tracking_number": tracking_info.get("@id"),
            "details": tracking_info.get("track_detail"),
            "summary": tracking_info.get("track_summary"),
        }

        return tracking_data

    def zip_to_city_state(self, zip_code):
        valid_address = {"city": "", "state": "XX", "zip_code": "XXXXX"}

        zip_code_dict = OrderedDict()
        zip_code_dict["@ID"] = "0"
        zip_code_dict["Zip5"] = zip_code

        city_state_request = OrderedDict()
        city_state_request["@USERID"] = self.usps_id
        city_state_request["ZipCode"] = zip_code_dict

        city_state_dict = {"CityStateLookupRequest": city_state_request}

        resp = self.usps_call(api="CityStateLookup", data=city_state_dict)

        if "zip_code" in resp:
            ret_zip = resp.get("zip_code")
            valid_address = {
                "city": ret_zip.get("city"),
                "state": ret_zip.get("state"),
                "zip_code": ret_zip.get("zip5"),
            }

        return valid_address



import datetime
import json
import logging
import os
import pprint

import requests
from future.standard_library import install_aliases

from geo_ez.data_functions import convert_keys

install_aliases()

logger = logging.getLogger(__name__)


class USCensus:
    is_debug = False
    is_logging = False

    today = None

    endpoint = "https://geocoding.geo.census.gov/geocoder"

    def __init__(self, **kwargs):
        self.is_debug = kwargs.get("debug", os.environ.get("FED_DEBUG", False)) in ["True", True]
        self.is_logging = kwargs.get("logging", os.environ.get("FED_LOGGING", False)) in ["True", True]

        self.today = datetime.datetime.now()

    def debug_msg(self, message, **kwargs):
        pretty = kwargs.get("pretty", False)

        if pretty:
            message = pprint.pformat(message, indent=4)

        if self.is_debug:
            print(message)

        if self.is_logging:
            logger.info(message)

    def geocode(self, **kwargs):
        return_address = None

        """
        - locations (to get just geocoding response)
        - geographies (to get geocoding response as well as geoLookup)
        """
        returntype = kwargs.get("returntype", "locations")

        # onelineaddress OR address OR coordinates
        searchtype = kwargs.get("searchtype", "address")

        """
         A numerical ID or name that references what version of the locator should be searched.
         - This generally corresponds to MTDB data which is benchmarked twice yearly.
         - A full list of options can be accessed at https://geocoding.geo.census.gov/geocoder/benchmarks.
         - The general format of the name is DatasetType_SpatialBenchmark.
         - The valid values for these include:
            - DatasetType
                - Public_AR
            - SpatialBenchmark
                - Current
                - ACS2018
                - Census2010

        So a resulting benchmark name could be 
            - Public_AR_Current
            - Public_AR_Census2010
            
        Over time, there will always be a 'Current' benchmark.
        It will change as the underlying dataset changes.
        """
        benchmark = kwargs.get("benchmark", "Public_AR_Current")

        """
        A numerical ID or name that references what vintage of geography is desired for the geoLookup. 
        (only needed when returntype = geographies)
        
        A full list of options for a given benchmark can be accessed at 
        https://geocoding.geo.census.gov/geocoder/vintages?benchmark=benchmarkId. 
        
        The general format of the name is GeographyVintage_SpatialBenchmark.
        - The SpatialBenchmark variable should always match the same named variable in what was chosen for 
          the benchmark parameter.
        - The GeographyVintage can be Current, ACS2018, etc.
        - So a resulting vintage name could be 'ACS2018_Current', 'Current_Census2010', etc.
        - Over time, there will always be a 'Current' vintage.
          It will change as the underlying dataset changes.
        """
        vintage = kwargs.get("vintage", "Current_Current")

        query = kwargs.get("query", False)

        if isinstance(query, dict):
            request_url_base = "%s/%s/%s" % (self.endpoint, returntype, searchtype)

            param_dict = dict(benchmark=benchmark, format="json")

            if returntype == "geographies":
                param_dict.update(dict(vintage=vintage))

            if searchtype == "onelineaddress":
                param_dict.update(dict(address=query))

            if searchtype == "coordinates":
                param_dict.update(dict(x=query.get("longitude"), y=query.get("latitude")))

            if searchtype == "address":
                param_dict.update(
                    dict(
                        street=query.get("address1"),
                        city=query.get("city"),
                        state=query.get("state"),
                        zip=query.get("zip_code"),
                    )
                )

            resp = requests.get(request_url_base, params=param_dict)

            if resp.status_code == 200:
                result = json.loads(resp.text).get("result")

                if isinstance(result, dict):
                    address_matches = result.get("addressMatches")
                    if len(address_matches) > 0:
                        address_match = address_matches[0]

                        return_address = convert_keys(address_match.get("addressComponents"))
                        return_address["latitude"] = address_match.get("coordinates").get("y")
                        return_address["longitude"] = address_match.get("coordinates").get("x")

        return return_address



from math import radians, cos, sin, asin, sqrt

from django.db import models
from django.db.models import DO_NOTHING

from geo_ez.constants import ACCURACY_CHOICES
from geo_ez.us_census_class import USCensus
from geo_ez.usps_class import USPS


class GISPoint(models.Model):
    name = models.CharField(max_length=180, blank=True, null=True)
    latitude = models.DecimalField(max_digits=22, decimal_places=16, blank=True, null=True)
    longitude = models.DecimalField(max_digits=22, decimal_places=16, blank=True, null=True)

    class Meta:
        abstract = True

    def __str__(self):
        return "%s,%s" % (str(self.latitude), str(self.longitude))

    def distance_from(self, latitude, longitude, **kwargs):
        use_miles = kwargs.get("use_miles", True)
        distance_unit = float(3959 if use_miles else 6371)

        latitude1 = radians(float(latitude))
        latitude2 = radians(float(self.latitude))

        longitude1 = radians(float(longitude))
        longitude2 = radians(float(self.longitude))

        distance_longitude = longitude2 - longitude1
        distance_latitude = latitude2 - latitude1

        a = sin(distance_latitude / 2) ** 2 + cos(latitude1) * cos(latitude2) * sin(distance_longitude / 2) ** 2
        c = 2 * asin(sqrt(a))

        distance = c * distance_unit

        return distance

    def in_radius(self, latitude, longitude, radius, **kwargs):
        return self.distance_from(latitude, longitude, **kwargs) < radius


class PostalCode(GISPoint):
    country_code = models.CharField(max_length=2, blank=True, null=True)  # iso country code, 2 characters
    postal_code = models.CharField(max_length=20, blank=True, null=True)  # varchar(20)
    place_name = models.CharField(max_length=180, blank=True, null=True)  # varchar(180)
    admin_name1 = models.CharField(max_length=100, blank=True, null=True)  # 1. order subdivision (state) varchar(100)
    admin_code1 = models.CharField(max_length=20, blank=True, null=True)  # 1. order subdivision (state) varchar(20)
    admin_name2 = models.CharField(
        max_length=100, blank=True, null=True
    )  # 2. order subdivision (county/province) varchar(100)
    admin_code2 = models.CharField(
        max_length=20, blank=True, null=True
    )  # 2. order subdivision (county/province) varchar(20)
    admin_name3 = models.CharField(
        max_length=100, blank=True, null=True
    )  # 3. order subdivision (community) varchar(100)
    admin_code3 = models.CharField(max_length=20, blank=True, null=True)  # 3. order subdivision (community) varchar(20)
    accuracy = models.IntegerField(
        null=True, choices=ACCURACY_CHOICES
    )  # accuracy of lat/lng from 1=estimated, 4=geonameid, 6=centroid of addresses or shape
    updated = models.DateTimeField(auto_now_add=True, null=True)

    def __str__(self):
        return self.place_name


class StreetAddress(GISPoint):
    address1 = models.CharField(max_length=255, blank=True, null=True)
    address2 = models.CharField(max_length=255, blank=True, null=True)
    city = models.CharField(max_length=180, blank=True, null=True)
    state = models.CharField(max_length=20, blank=True, null=True)
    zip_code = models.CharField(max_length=20, blank=True, null=True)
    plus_four = models.CharField(max_length=20, blank=True, null=True)
    postal_code = models.ForeignKey(PostalCode, blank=True, null=True, on_delete=DO_NOTHING)
    validated = models.BooleanField(default=False)

    def dict(self):
        return dict(
            address1=self.address1, address2=self.address2, city=self.city, state=self.state, zip_code=self.zip_code
        )

    def geocode(self, **kwargs):
        retn = False

        valid_address = self.normalize()

        usc = USCensus()

        geocoded = usc.geocode(query=valid_address)
        if geocoded:
            retn = True
            self.latitude = geocoded.get("latitude")
            self.longitude = geocoded.get("longitude")

        return retn

    def link_postal_code(self):
        retn = False
        try:
            self.postal_code = PostalCode.objects.get(postal_code=self.zip_code)

        except PostalCode.DoesNotExist:
            pass

        else:
            retn = True
            self.save()

        return retn

    def normalize(self):
        valid_address = dict(
            address1=self.address1,
            address2=self.address2,
            city=self.city,
            state=self.state,
            zip_code=self.zip_code,
            plus_four=self.plus_four,
        )

        if not self.validated:
            search_address = self.dict()

            ps = USPS()
            valid_address = ps.address(**search_address)

            self.address1 = valid_address.get("address1", self.address1)
            self.address2 = valid_address.get("address2", self.address2)
            self.city = valid_address.get("city", self.city)
            self.state = valid_address.get("state", self.state)
            self.zip_code = valid_address.get("zip_code", self.zip_code)
            self.plus_four = valid_address.get("plus_four", self.plus_four)
            self.validated = True
            self.save()

        return valid_address



# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import PostalCode, StreetAddress


@admin.register(PostalCode)
class PostalCodeAdmin(admin.ModelAdmin):
    list_display = (
        "country_code",
        "postal_code",
        "place_name",
        "admin_name1",
        "admin_code1",
        "admin_name2",
        "admin_code2",
        "admin_name3",
        "admin_code3",
        "latitude",
        "longitude",
        "accuracy",
    )
    list_filter = ("admin_name1",)
    search_fields = ("place_name", "postal_code", "admin_name1", "admin_code1", "admin_name2")


@admin.register(StreetAddress)
class StreetAddressAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "address1",
        "address2",
        "city",
        "state",
        "zip_code",
    )
    list_filter = ("postal_code", "state", "validated")
    search_fields = ("name", "city", "state", "zip_code")

    def save_model(self, request, obj, form, change):
        super(StreetAddressAdmin, self).save_model(request, obj, form, change)

        obj.link_postal_code()
        obj.normalize()
        obj.geocode()



import csv
import datetime

from django.db.models.expressions import RawSQL
from django.utils.timezone import make_aware

from geo_ez.models import PostalCode


def points_within_radius(gismodel, latitude, longitude, **kwargs):
    radius = kwargs.get("radius", False)
    use_miles = kwargs.get("use_miles", True)

    if radius:
        radius = float(radius)

    distance_unit = float(3959 if use_miles else 6371)

    # Great circle distance formula
    gcd_formula = (
        "%s * acos(least(greatest(cos(radians(%s)) * cos(radians(latitude)) * cos(radians(longitude) - "
        "radians(%s)) + sin(radians(%s)) * sin(radians(latitude)), -1), 1))"
    )
    distance_raw_sql = RawSQL(gcd_formula, (distance_unit, latitude, longitude, latitude))
    qs = gismodel.objects.all().annotate(distance=distance_raw_sql).order_by("distance")

    if radius:
        qs = qs.filter(distance__lt=radius)

    return qs


def postal_codes_within_radius(latitude, longitude, **kwargs):
    return points_within_radius(PostalCode, latitude, longitude, **kwargs)


def import_postal_codes_csv(data_file_path, **kwargs):
    delimiter = kwargs.get("delimiter", "\t")

    data_file = open(data_file_path, "rU", encoding="utf-8")

    rows = csv.reader(data_file, delimiter=delimiter)

    insert_list = []
    for row in rows:
        if len(row) > 0 and row[11]:
            try:
                postal_code = PostalCode.objects.get(postal_code=row[1], name=row[2], place_name=row[2])

            except PostalCode.DoesNotExist:
                insert_list.append(
                    PostalCode(
                        country_code=row[0],
                        postal_code=row[1],
                        name=row[2],
                        place_name=row[2],
                        admin_name1=row[3],
                        admin_code1=row[4],
                        admin_name2=row[5],
                        admin_code2=row[6],
                        admin_name3=row[7],
                        admin_code3=row[8],
                        latitude=row[9],
                        longitude=row[10],
                        accuracy=row[11],
                        updated=make_aware(datetime.datetime.now()),
                    )
                )

            else:
                postal_code.country_code = row[0]
                postal_code.postal_code = row[1]
                postal_code.name = row[2]
                postal_code.place_name = row[2]
                postal_code.admin_name1 = row[3]
                postal_code.admin_code1 = row[4]
                postal_code.admin_name2 = row[5]
                postal_code.admin_code2 = row[6]
                postal_code.admin_name3 = row[7]
                postal_code.admin_code3 = row[8]
                postal_code.latitude = row[9]
                postal_code.longitude = row[10]
                postal_code.accuracy = row[11]
                postal_code.updated = make_aware(datetime.datetime.now())
                postal_code.save()

    data_file.close()

    PostalCode.objects.bulk_create(insert_list)



import re
from os import path

from setuptools import setup

github_user = "avryhof"
github_repo_name = "geo_ez"
module_path = "geo_ez"
author = "Amos Vryhof"
author_email = "amos@vryhofresearch.com"
description = "Tools for building geographically aware django apps, without needing to install a real GIS backend."


def read(*parts):
    return open(path.join(path.dirname(__file__), *parts)).read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def find_name(*file_paths):
    name_file = read(*file_paths)
    name_match = re.search(r"^__name__ = ['\"]([^'\"]*)['\"]", name_file, re.M)
    if name_match:
        return name_match.group(1)
    return module_path


setup(
    name=find_name(module_path, "__init__.py"),
    version=find_version(module_path, "__init__.py"),
    packages=[find_name(module_path, "__init__.py")],
    url="https://github.com/%s/%s" % (github_user, github_repo_name),
    license="MIT",
    author=author,
    author_email=author_email,
    description=description,
    long_description="",
    include_package_data=True,
    package_data={"": ["LICENSE"]},
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    install_requires=["bleach", "django", "future", "requests", "xmltodict"],
)
